"""
Contains common exceptions used by ttast
"""

class PipelineRunException(Exception):
    """
    Exception representing a runtime error while processing the pipeline
    """
    pass

class PipelineConfigException(Exception):
    """
    Exception representing a config error while processing the pipeline
    """
    pass

class ValidationException(Exception):
    """
    Validation of some condition failed
    """




import os
import copy
import inspect

from .util import *
from .exception import *
from . import types
from . import builtin

logger = logging.getLogger(__name__)

def builtin_handlers():

    return {
        "config": builtin.HandlerConfig,
        "import": builtin.HandlerImport,
        "meta": builtin.HandlerMeta,
        "replace": builtin.HandlerReplace,
        "split_yaml": builtin.HandlerSplitYaml,
        "stdin": builtin.HandlerStdin,
        "stdout": builtin.HandlerStdout,
        "template": builtin.HandlerTemplate,
        "sum": builtin.HandlerSum
    }

def builtin_support_handlers():

    return [
        builtin.SupportHandlerMatchTags,
        builtin.SupportHandlerWhen,
        builtin.SupportHandlerTags
    ]

def run_pipeline(pipeline_steps, include_builtin=True, custom_handlers=None, custom_support_handlers=None):
    validate(isinstance(pipeline_steps, list) and all(isinstance(x, dict) for x in pipeline_steps),
        "Pipeline steps passed to run_pipeline must be a list of dictionaries")
    validate(isinstance(include_builtin, bool), "Invalid include_builtin parameter passed to run_pipeline")
    validate(isinstance(custom_handlers, dict) or custom_handlers is None, "Invalid custom_handlers passed to run_pipeline")
    validate(isinstance(custom_support_handlers, list) or custom_support_handlers is None, "Invalid custom_support_handlers passed to run_pipeline")

    pipeline = types.Pipeline()

    if include_builtin:
        pipeline.add_handlers(builtin_handlers())
        pipeline.add_support_handlers(builtin_support_handlers())

    if custom_handlers is not None:
        pipeline.add_handlers(custom_handlers)

    if custom_support_handlers is not None:
        pipeline.add_support_handlers(custom_support_handlers)

    for step in pipeline_steps:
        pipeline.add_step(step)

    # Run the pipeline to completion
    pipeline.run()



#!/usr/bin/env python3
"""
"""

import argparse
import logging
import sys

from .exception import *
from .util import *
from . import pipeline

logger = logging.getLogger(__name__)

def process_args() -> int:
    """
    Processes ttast command line arguments, initialises and runs the pipeline to perform text processing
    """

    # Create parser for command line arguments
    parser = argparse.ArgumentParser(
        prog="ttast", description="Text Transform Assistant", exit_on_error=False
    )

    # Parser configuration
    parser.add_argument(
        "-c", action="append", dest="configs", help="Configuration files"
    )

    parser.add_argument(
        "-d", action="store_true", dest="debug", help="Enable debug output"
    )

    args = parser.parse_args()

    # Capture argument options
    debug = args.debug
    configs = args.configs

    # Logging configuration
    level = logging.WARNING
    if debug:
        level = logging.DEBUG

    logging.basicConfig(level=level, format="%(levelname)s: %(message)s")

    try:
        # Add each config as a pipeline step, which will read and merge the config
        steps = list()
        if configs is not None:
            for config_item in configs:
                # If '-' is specified, read the configuration from stdin
                if config_item == "-":
                    step_def = {
                        "type": "config",
                        "stdin": True
                    }
                else:
                    step_def = {
                        "type": "config",
                        "file": config_item
                    }

                steps.append(step_def)

        pipeline.run_pipeline(steps)

    except Exception as e:  # pylint: disable=broad-exception-caught
        if debug:
            logger.error(e, exc_info=True, stack_info=True)
        else:
            logger.error(e)
        return 1

    return 0


def main():
    """
    Entrypoint for the module.
    Minor exception handling is performed, along with return code processing and
    flushing of stdout on program exit.
    """
    try:
        ret = process_args()
        sys.stdout.flush()
        sys.exit(ret)
    except Exception as e:  # pylint: disable=broad-exception-caught
        logging.getLogger(__name__).exception(e)
        sys.stdout.flush()
        sys.exit(1)


if __name__ == "__main__":
    main()




import yaml
import sys
import re
import glob
import os
import inspect
import copy
import logging

from .util import *

logger = logging.getLogger(__name__)

class TextBlock:
    def __init__(self, text, *, tags=None):
        validate(isinstance(tags, (list, set)) or tags is None, "Tags supplied to TextBlock must be a set, list or absent")

        self.text = text

        self.tags = set()
        if tags is not None:
            for tag in tags:
                self.tags.add(tag)

        self.meta = {}

class Pipeline:
    def __init__(self):
        self._steps = []
        self._handlers = {}
        self._support_handlers = []
        self._vars = {
            "env": os.environ.copy()
        }
        self._blocks = []

        self.step_limit = 100

    def add_block(self, block):
        validate(isinstance(block, TextBlock), "Invalid block passed to pipeline add_block")

        self._blocks.append(block)

    def remove_block(self, block):
        validate(isinstance(block, TextBlock), "Invalid block passed to pipeline remove_block")

        self._blocks.remove(block)

    def copy_vars(self):
        return copy.deepcopy(self._vars)

    def set_var(self, key, value):
        if key in ["env"]:
            raise PipelineRunException(f"Disallowed key in pipeline set_var: {key}")

        self._vars[key] = value

    def add_step(self, step_def):
        validate(isinstance(step_def, dict), "Invalid step definition passed to add_step")

        if self.step_limit > 0 and len(self._steps) > self.step_limit:
            raise PipelineRunException(f"Reached limit of {self.step_limit} steps in pipeline. This is a safe guard to prevent infinite recursion")

        self._steps.append(step_def)

    def add_handlers(self, handlers):
        validate(isinstance(handlers, dict), "Invalid handlers passed to add_handlers")
        validate((all(x is None or (inspect.isclass(x) and issubclass(x, Handler))) for x in handlers.values()), "Invalid handlers passed to add_handlers")

        for key in handlers:
            self._handlers[key] = handlers[key]

    def add_support_handlers(self, handlers):
        validate(isinstance(handlers, list), "Invalid handlers passed to add_support_handlers")
        validate((all(inspect.isclass(x) and issubclass(x, SupportHandler)) for x in handlers), "Invalid handlers passed to add_support_handlers")

        for handler in handlers:
            if handler not in self._support_handlers:
                self._support_handlers.append(handler)

    def run(self):
        # This is a while loop with index to allow the pipeline to be appended to during processing
        index = 0
        while index < len(self._steps):

            # Clone current step definition
            step_def = self._steps[index].copy()
            index = index + 1

            # Extract type
            step_type = pop_property(step_def, "type", template_map=None)
            validate(isinstance(step_type, str) and step_type != "", "Step 'type' is required and must be a non empty string")

            # Retrieve the handler for the step type
            handler = self._handlers.get(step_type)
            if handler is None:
                raise PipelineRunException(f"Invalid step type in step {step_type}")

            # Create an instance per block for the step type, or a single instance for step types
            # that are not per block.
            if handler.is_per_block():
                logger.debug(f"Processing {step_type} - per_block")
                # Create a copy of blocks to allow steps to alter the block list while we iterate
                block_list_copy = self._blocks.copy()

                for block in block_list_copy:
                    self._process_step_instance(step_def, handler, block)
            else:
                logger.debug(f"Processing {step_type} - singular")
                self._process_step_instance(step_def, handler)

    def _process_step_instance(self, step_def, handler, block=None):
        validate(isinstance(step_def, dict), "Invalid step definition passed to _process_step_instance")
        validate(inspect.isclass(handler) and issubclass(handler, Handler), "Invalid handler passed to _process_step_instance")
        validate(block is None or isinstance(block, TextBlock), "Invalid text block passed to _process_step_instance")

        # Create new vars for the instance, based on the pipeline vars, plus including
        # any block vars, if present
        step_vars = self.copy_vars()
        if block is not None:
            step_vars["meta"] = copy.deepcopy(block.meta)
            step_vars["tags"] = copy.deepcopy(block.tags)

        state = PipelineStepState(step_def, self, step_vars)

        #
        # Parsing
        #

        # Initialise and parse support handlers
        support_handlers = [x() for x in self._support_handlers]
        for support in support_handlers:
            support.init(state)
            support.parse()

        # Initialise and parse the main handler
        instance = handler()
        instance.init(state)
        instance.parse()

        # At this point, there should be no properties left in the dictionary as all of the handlers should have
        # extracted their own properties.
        validate(len(state.step_def.keys()) == 0, f"Unknown properties on step definition: {list(state.step_def.keys())}")

        #
        # Execution
        #

        # Processing may start with a single block, which may become more, none or stay at one
        # block.
        # If a block is split, then two blocks replace it, while a block may disappear from the list
        # if the handler filters it out.
        # Any new blocks should also be processed by the following handlers, so the working block
        # list is dynamic.

        # block_list contains a list of the blocks to operate on for the current handler, while working_list
        # is the list being generated for the next handler to operate on.

        block_list = [block]
        working_list = []

        if block is not None:
            logger.debug(f"Operating on block: {hex(id(block))}")
            logger.debug(f" meta: {block.meta}")
            logger.debug(f" tags: {block.tags}")

        # Run any preprocessing handlers
        for support in support_handlers:
            for current_block in block_list:
                logger.debug(f"Calling pre support handler ({support}) for block {hex(id(current_block))}")
                result = support.pre(current_block)
                if result is None:
                    # If the handler didn't return anything, then just add the current block
                    # for the next round of handlers
                    working_list.append(current_block)
                else:
                    # The handler returned a replacement list of blocks, which should replace the current
                    # block in the block list for the next round of handlers.
                    # This list could also be empty, removing the block from further processing
                    for x in result:
                        working_list.append(x)

            # Update block_list with the new list of blocks for the next round of handlers and reset
            # working_list
            block_list = working_list
            working_list = []

        # Perform processing for the main handler
        for current_block in block_list:
            logger.debug(f"Calling handler ({instance}) for block {hex(id(current_block))}")
            result = instance.run(current_block)
            if result is None:
                working_list.append(current_block)
            else:
                for x in result:
                    working_list.append(x)

        block_list = working_list
        working_list = []

        # Run any post processing handlers
        for support in support_handlers:
            for current_block in block_list:
                logger.debug(f"Calling post support handler ({support}) for block {hex(id(current_block))}")
                result = support.post(current_block)
                if result is None:
                    working_list.append(current_block)
                else:
                    for x in result:
                        working_list.append(x)

            block_list = working_list
            working_list = []

class PipelineStepState:
    def __init__(self, step_def, pipeline, step_vars):
        validate(isinstance(step_def, dict), "Invalid step_def passed to PipelineStepState")
        validate(isinstance(pipeline, Pipeline) or pipeline is None, "Invalid pipeline passed to PipelineStepState")
        validate(isinstance(step_vars, dict), "Invalid step vars passed to PipelineStepState")

        self.step_def = step_def.copy()
        self.pipeline = pipeline
        self.vars = step_vars

class SupportHandler:
    def init(self, state):
        validate(isinstance(state, PipelineStepState), "Invalid step state passed to SupportHandler")

        self.state = state

    def parse(self):
        raise PipelineRunException("parse undefined in SupportHandler")

    def pre(self, block):
        raise PipelineRunException("pre undefined in SupportHandler")

    def post(self, block):
        raise PipelineRunException("post undefined in SupportHandler")

class Handler:
    def is_per_block(self):
        raise PipelineRunException("is_per_block undefined in Handler")

    def init(self, state):
        validate(isinstance(state, PipelineStepState), "Invalid step state passed to Handler")

        self.state = state

    def parse(self):
        raise PipelineRunException("parse undefined in Handler")

    def run(self, block):
        raise PipelineRunException("run undefined in Handler")




import jinja2
import logging

from .exception import *

logger = logging.getLogger(__name__)

def validate(val, message):
    if not val:
        raise ValidationException(message)

def template_if_string(val, mapping):
    if val is not None and isinstance(val, str):
        try:
            environment = jinja2.Environment()

            template = environment.from_string(val)
            return template.render(mapping)
        except KeyError as e:
            raise PipelineRunException(f"Missing key in template substitution: {e}") from e

    return val

def parse_bool(obj) -> bool:
    if obj is None:
        raise PipelineRunException("None value passed to parse_bool")

    if isinstance(obj, bool):
        return obj

    obj = str(obj)

    if obj.lower() in ["true", "1"]:
        return True

    if obj.lower() in ["false", "0"]:
        return False

    raise PipelineRunException(f"Unparseable value ({obj}) passed to parse_bool")

def pop_property(spec, key, *, template_map=None, default=None, required=False):
    validate(isinstance(spec, dict), f"Invalid spec passed to pop_property. Must be dict")
    validate(isinstance(template_map, dict) or template_map is None, "Invalid type passed as template_map")

    if key not in spec:
        # Raise exception is the key isn't present, but required
        if required:
            raise KeyError(f'Missing key "{key}" in spec or value is null')

        # If the key is not present, return the default
        return default

    # Retrieve value
    val = spec.pop(key)

    # Template the value, depending on the type
    if val is not None and template_map is not None:
        if isinstance(val, str):
            val = template_if_string(val, template_map)
        elif isinstance(val, list):
            val = [template_if_string(x, template_map) for x in val]
        elif isinstance(val, dict):
            for val_key in val:
                val[val_key] = template_if_string(val[val_key], template_map)

    return val

def merge_meta_tags(vars, *, tags=None, meta=None):
    validate(isinstance(vars, dict), "Vars provided to merge_meta_tags is not a dictionary")
    validate(isinstance(tags, (set, list)) or tags is None, "Tags provided to merge_meta_tags is not a list, set or absent")
    validate(isinstance(meta, dict) or tags is None, "Tags provided to merge_meta_tags is not a list or absent")

    new_vars = vars.copy()

    new_tags = ",".join(set(tags))
    new_vars["ttast_tags"] = new_tags

    # Create vars for all of the metadata
    for meta_key in meta:
        new_vars[f"ttast_meta_{meta_key}"] = meta[meta_key]

    return new_vars



from .ttast import main

__all__ = ["main"]

from .pipeline import run_pipeline
from .pipeline import builtin_support_handlers
from .pipeline import builtin_handlers

from .types import Pipeline
from .types import TextBlock
from .types import PipelineStepState
from .types import SupportHandler
from .types import Handler



import os

from setuptools import find_packages, setup

with open("README.md") as readme_file:
    README = readme_file.read()

setup_args = {
    "name": "ttast",
    "version": os.environ["BUILD_VERSION"],
    "description": "Text Transform Assistant",
    "long_description_content_type": "text/markdown",
    "long_description": README,
    "license": "MIT",
    "packages": find_packages(where="src", include=["ttast", "ttast.*"]),
    "author": "Jesse Reichman",
    "keywords": ["Text", "Transform", "Assistant"],
    "url": "https://github.com/archmachina/ttast",
    "download_url": "https://pypi.org/project/ttast/",
    "entry_points": {"console_scripts": ["ttast = ttast:main"]},
    "package_dir": {"": "src"},
    "install_requires": ["PyYAML>=6.0.0", "Jinja2>=3.1.0"],
}

if __name__ == "__main__":
    setup(**setup_args, include_package_data=True)




import yaml
import sys
import re
import glob
import os
import inspect
import copy
import hashlib
import textwrap

from .util import *
from . import types

class HandlerConfig(types.Handler):
    """
    """
    def parse(self):
        # Read the content from the file and use _process_config_content to do the work
        self.config_file = pop_property(self.state.step_def, "file", template_map=self.state.vars)
        validate(isinstance(self.config_file, str) or self.config_file is None, "Step 'config_file' must be a string or absent")
        validate(not isinstance(self.config_file, str) or self.config_file != "", "Step 'config_file' cannot be empty")

        # Extract the content var, which can be either a dict or yaml string
        self.config_content = pop_property(self.state.step_def, "content", template_map=self.state.vars)
        validate(isinstance(self.config_content, (str, dict)) or self.config_content is None, "Step 'config_content' must be a string, dict or absent")

        # Extract stdin bool, indicating whether to read config from stdin
        self.stdin = pop_property(self.state.step_def, "stdin", template_map=self.state.vars, default=False)
        validate(isinstance(self.stdin, (bool, str)), "Step 'stdin' must be a bool, bool like string or absent")
        self.stdin = parse_bool(self.stdin)

    def is_per_block():
        return False

    def run(self, block):
        if self.config_file is not None:
            logger.debug(f"config: including config from file {self.config_file}")
            with open(self.config_file, "r", encoding='utf-8') as file:
                content = file.read()

            self._process_config_content(content)

        # Call _process_config_content, which can determine whether to process as string or dict
        if self.config_content is not None:
            logger.debug(f"config: including inline config")
            self._process_config_content(self.config_content)

        if self.stdin:
            # Read configuration from stdin
            logger.debug(f"config: including stdin config")
            stdin_content = sys.stdin.read()
            self._process_config_content(stdin_content)

    def _process_config_content(self, content):
        validate(isinstance(content, (str, dict)), "Included configuration must be a string or dictionary")

        # Don't error on an empty configuration. Just return
        if content == "":
            logger.debug("config: empty configuration. Ignoring.")
            return

        # Parse yaml if it is a string
        if isinstance(content, str):
            content = yaml.safe_load(content)

        validate(isinstance(content, dict), "Parsed configuration is not a dictionary")

        # Extract vars from the config
        config_vars = pop_property(content, "vars", template_map=self.state.vars, default={})
        validate(isinstance(config_vars, dict), "Config 'vars' is not a dictionary")

        for config_var_name in config_vars:
            self.state.pipeline.set_var(config_var_name, config_vars[config_var_name])

        # Extract pipeline steps from the config
        config_pipeline = pop_property(content, "pipeline", template_map=None, default=[])
        validate(isinstance(config_pipeline, list), "Config 'pipeline' is not a list")

        for step in config_pipeline:
            validate(isinstance(step, dict), "Pipeline entry is not a dictionary")

            self.state.pipeline.add_step(step)

        # Validate config has no other properties
        validate(len(content.keys()) == 0, f"Found unknown properties in configuration: {content.keys()}")

class HandlerImport(types.Handler):
    """
    """
    def parse(self):
        self.import_files = pop_property(self.state.step_def, "files", template_map=self.state.vars)
        validate(isinstance(self.import_files, list), "Step 'files' must be a list of strings")
        validate(all(isinstance(x, str) for x in self.import_files), "Step 'files' must be a list of strings")

        self.recursive = pop_property(self.state.step_def, "recursive", template_map=self.state.vars)
        validate(isinstance(self.recursive, (bool, str)), "Step 'recursive' must be a bool or bool like string")
        self.recursive = parse_bool(self.recursive)

    def is_per_block():
        return False

    def run(self, block):
        filenames = set()
        for import_file in self.import_files:
            logger.debug(f"import: processing file glob: {import_file}")
            matches = glob.glob(import_file, recursive=self.recursive)
            for match in matches:
                filenames.add(match)

        # Ensure consistency for load order
        filenames = list(filenames)
        filenames.sort()

        new_blocks = []
        for filename in filenames:
            logger.debug(f"import: reading file {filename}")
            with open(filename, "r", encoding="utf-8") as file:
                content = file.read()
                new_block = types.TextBlock(content)
                new_block.meta["import_filename"] = filename
                self.state.pipeline.add_block(new_block)
                new_blocks.append(new_block)

        return new_blocks

class HandlerMeta(types.Handler):
    """
    """
    def parse(self):
        self.vars = pop_property(self.state.step_def, "vars", template_map=self.state.vars)
        validate(isinstance(self.vars, dict), "Step 'vars' must be a dictionary of strings")
        validate(all(isinstance(x, str) for x in self.vars), "Step 'vars' must be a dictionary of strings")

    def is_per_block():
        return True

    def run(self, block):
        for key in self.vars:
            block.meta[key] = self.vars[key]

class HandlerReplace(types.Handler):
    """
    """
    def parse(self):
        self.replace = pop_property(self.state.step_def, "replace", template_map=self.state.vars, default={})
        validate(isinstance(self.replace, list), "Step 'replace' must be a list")
        validate(all(isinstance(x, dict) for x in self.replace), "Step 'replace' items must be dictionaries")
        for item in self.replace:
            validate('key' in item and isinstance(item['key'], str), "Step 'replace' items must contain a string 'key' property")
            validate('value' in item and isinstance(item['value'], str), "Step 'replace' items must contain a string 'value' property")

        self.regex = pop_property(self.state.step_def, "regex", template_map=self.state.vars, default=False)
        validate(isinstance(self.regex, (bool, str)), "Step 'regex' must be a bool, bool like string or absent")
        self.regex = parse_bool(self.regex)

    def is_per_block():
        return True

    def run(self, block):
        for replace_item in self.replace:
            # Copy the dictionary as we'll change it when removing values
            replace_item = replace_item.copy()

            replace_key = replace_item['key']
            replace_value = replace_item['value']

            replace_regex = pop_property(replace_item, "regex", template_map=self.state.vars, default=False)
            validate(isinstance(replace_regex, (bool, str)), "Replace item 'regex' must be a bool, bool like string or absent")
            replace_regex = parse_bool(replace_regex)

            # replace_value isn't templated by pop_property as it is a list of dictionaries, so it
            # needs to be manually done here
            replace_value = template_if_string(replace_value, self.state.vars)

            logger.debug(f"replace: replacing regex({self.regex or replace_regex}): {replace_key} -> {replace_value}")

            if self.regex or replace_regex:
                block.text = re.sub(replace_key, replace_value, block.text)
            else:
                block.text = block.text.replace(replace_key, replace_value)

class HandlerSplitYaml(types.Handler):
    """
    """
    def parse(self):
        self.strip = pop_property(self.state.step_def, "strip", template_map=self.state.vars, default=False)
        validate(isinstance(self.strip, (bool, str)), "Step 'strip' must be a bool or str value")
        self.strip = parse_bool(self.strip)

    def is_per_block():
        return True

    def run(self, block):
        lines = block.text.splitlines()
        documents = []
        current = []

        for line in lines:

            # Determine if we have the beginning of a yaml document
            if line == "---" and len(current) > 0:
                documents.append("\n".join(current))
                current = []

            current.append(line)

        documents.append("\n".join(current))

        # Strip each document, if required
        if self.strip:
            documents = [x.strip() for x in documents]

        # If we have a single document and it's the same as the
        # original block, just exit
        if len(documents) == 1 and documents[0] == block.text:
            return

        # Add all documents to the pipeline text block list
        new_blocks = [types.TextBlock(item) for item in documents]
        for new_block in new_blocks:
            new_block.meta = copy.deepcopy(block.meta)
            new_block.tags = copy.deepcopy(block.tags)

            self.state.pipeline.add_block(new_block)

        # Remove the original source block from the list
        self.state.pipeline.remove_block(block)

        logger.debug(f"split_yaml: output 1 document -> {len(documents)} documents")

        return new_blocks

class HandlerStdin(types.Handler):
    """
    """
    def parse(self):
        self.split = pop_property(self.state.step_def, "split", template_map=self.state.vars)
        validate(isinstance(self.split, str) or self.split is None, "Step 'split' must be a string")

        self.strip = pop_property(self.state.step_def, "strip", template_map=self.state.vars, default=False)
        validate(isinstance(self.strip, (bool, str)), "Step 'strip' must be a bool or str value")
        self.strip = parse_bool(self.strip)

    def is_per_block():
        return False

    def run(self, block):
        # Read content from stdin
        logger.debug("stdin: reading document from stdin")
        stdin_content = sys.stdin.read()

        # Split if required and convert to a list of documents
        if self.split is not None and self.split != "":
            stdin_items = stdin_content.split(self.split)
        else:
            stdin_items = [stdin_content]

        # strip leading and trailing whitespace, if required
        if self.strip:
            stdin_items = [x.strip() for x in stdin_items]

        # Add the stdin items to the list of text blocks
        new_blocks = [types.TextBlock(item) for item in stdin_items]
        for item in new_blocks:
            self.state.pipeline.add_block(item)

        return new_blocks

class HandlerStdout(types.Handler):
    """
    """
    def parse(self):
        self.prefix = pop_property(self.state.step_def, "prefix", template_map=self.state.vars)
        validate(isinstance(self.prefix, str) or self.prefix is None, "Step 'prefix' must be a string")

        self.suffix = pop_property(self.state.step_def, "suffix", template_map=self.state.vars)
        validate(isinstance(self.suffix, str) or self.suffix is None, "Step 'suffix' must be a string")

    def is_per_block():
        return True

    def run(self, block):
        if self.prefix is not None:
            print(self.prefix)

        print(block.text)

        if self.suffix is not None:
            print(self.suffix)

class HandlerTemplate(types.Handler):
    """
    """
    def parse(self):
        self.vars = pop_property(self.state.step_def, "vars", template_map=self.state.vars)
        validate(isinstance(self.vars, dict) or self.vars is None, "Step 'vars' must be a dictionary or absent")

    def is_per_block():
        return True

    def run(self, block):
        template_vars = self.state.vars.copy()

        if self.vars is not None:
            for key in self.vars:
                template_vars[key] = self.vars[key]

        environment = jinja2.Environment()
        template = environment.from_string(block.text)
        block.text = template.render(template_vars)

def _block_sum(block):
    validate(isinstance(block, types.TextBlock), "Invalid text block passed to _block_sum")

    md5 = hashlib.md5()
    sha1 = hashlib.sha1()
    sha256 = hashlib.sha256()

    encode = block.text.encode("utf-8")
    md5.update(encode)
    sha1.update(encode)
    sha256.update(encode)

    block.meta["md5sum"] = md5.hexdigest()
    block.meta["sha1sum"] = sha1.hexdigest()
    block.meta["sha256sum"] = sha256.hexdigest()

    short = 0
    short_wrap = textwrap.wrap(sha256.hexdigest(), 8)
    for item in short_wrap:
        short = short ^ int(item, 16)

    block.meta["shortsum"] = format(short, 'x')

class HandlerSum(types.Handler):
    """
    """
    def parse(self):
        pass

    def is_per_block():
        return True

    def run(self, block):
        _block_sum(block)

        logger.debug(f"sum: document short sum: {block.meta['shortsum']}")

class SupportHandlerSum(types.SupportHandler):
    """
    """
    def parse(self):
        pass

    def pre(self, block):
        pass

    def post(self, block):
        if block is None:
            return

        _block_sum(block)

        logger.debug(f"sum: document short sum: {block.meta['shortsum']}")

class SupportHandlerTags(types.SupportHandler):
    def parse(self):
        # Apply tags
        self.apply_tags = pop_property(self.state.step_def, "apply_tags", template_map=self.state.vars, default=[])
        validate(isinstance(self.apply_tags, list), "Step 'apply_tags' must be a list of strings")
        validate(all(isinstance(x, str) for x in self.apply_tags), "Step 'apply_tags' must be a list of strings")

    def pre(self, block):
        pass

    def post(self, block):
        if block is not None:
            for tag in self.apply_tags:
                block.tags.add(tag)

class SupportHandlerWhen(types.SupportHandler):
    def parse(self):
        # When condition
        self.when = pop_property(self.state.step_def, "when", template_map=self.state.vars, default=[])
        validate(isinstance(self.when, (list, str)), "Step 'when' must be a string or list of strings")
        if isinstance(self.when, str):
            self.when = [self.when]
        validate(all(isinstance(x, str) for x in self.when), "Step 'when' must be a string or list of strings")

    def pre(self, block):
        if len(self.when) > 0:
            environment = jinja2.Environment()
            for condition in self.when:
                template = environment.from_string("{{" + condition + "}}")
                if not parse_bool(template.render(self.state.vars)):
                    return []

    def post(self, block):
        pass

class SupportHandlerMatchTags(types.SupportHandler):
    def parse(self):
        # Extract match any tags
        self.match_any_tags = pop_property(self.state.step_def, "match_any_tags", template_map=self.state.vars, default=[])
        validate(isinstance(self.match_any_tags, list), "Step 'match_any_tags' must be a list of strings")
        validate(all(isinstance(x, str) for x in self.match_any_tags), "Step 'match_any_tags' must be a list of strings")
        self.match_any_tags = set(self.match_any_tags)

        # Extract match all tags
        match_all_tags = pop_property(self.state.step_def, "match_all_tags", template_map=self.state.vars, default=[])
        validate(isinstance(match_all_tags, list), "Step 'match_all_tags' must be a list of strings")
        validate(all(isinstance(x, str) for x in match_all_tags), "Step 'match_all_tags' must be a list of strings")
        self.match_all_tags = set(match_all_tags)

        # Extract exclude tags
        self.exclude_tags = pop_property(self.state.step_def, "exclude_tags", template_map=self.state.vars, default=[])
        validate(isinstance(self.exclude_tags, list), "Step 'exclude_tags' must be a list of strings")
        validate(all(isinstance(x, str) for x in self.exclude_tags), "Step 'exclude_tags' must be a list of strings")
        self.exclude_tags = set(self.exclude_tags)

    def pre(self, block):
        if len(self.match_any_tags) > 0:
            # If there are any 'match_any_tags', then at least one of them has to match with the document
            if len(self.match_any_tags.intersection(block.tags)) == 0:
                return []

        if len(self.match_all_tags) > 0:
            # If there are any 'match_all_tags', then all of those tags must match the document
            for tag in self.match_all_tags:
                if tag not in block.tags:
                    return []

        if len(self.exclude_tags) > 0:
            # If there are any exclude tags and any are present in the block, it isn't a match
            for tag in self.exclude_tags:
                if tag in block.tags:
                    return []

    def post(self, block):
        pass
